/*
 * Copyright (C) 2003  MaxSt ( maxst@hiend3d.com )
 *
 * Copyright (C) 2015  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include "hqx-format.h"
#include "hqx-common.h"

#if ( \
    defined(__ARM_ARCH_6__) || \
    defined(__ARM_ARCH_6J__) || \
    defined(__ARM_ARCH_6K__) || \
    defined(__ARM_ARCH_6Z__) || \
    defined(__ARM_ARCH_6ZK__) || \
    defined(__ARM_ARCH_6T2__) || \
    defined(__ARM_ARCH_7__) || \
    defined(__ARM_ARCH_7A__) || \
    defined(__ARM_ARCH_7R__) || \
    defined(__ARM_ARCH_7M__) || \
    defined(__ARM_ARCH_7S__) || \
    (defined(_M_ARM) && (_M_ARM >= 6)) || \
    (defined(__TARGET_ARCH_ARM) && (__TARGET_ARCH_ARM >= 6)) || \
    (defined(__TARGET_ARCH_THUMB) && (__TARGET_ARCH_THUMB >= 3)) \
)
    #define ARMV6 1
#else
    #undef ARMV6
#endif

#if ( \
    defined(__ARM_ARCH_7__) || \
    defined(__ARM_ARCH_7A__) || \
    defined(__ARM_ARCH_7R__) || \
    defined(__ARM_ARCH_7M__) || \
    defined(__ARM_ARCH_7S__) || \
    (defined(_M_ARM) && (_M_ARM >= 7)) || \
    (defined(__TARGET_ARCH_ARM) && (__TARGET_ARCH_ARM >= 7)) || \
    (defined(__TARGET_ARCH_THUMB) && (__TARGET_ARCH_THUMB >= 4)) \
)
    #define ARMV7 1
#else
    #undef ARMV7
#endif



#define Ymask 0x00FF0000
#define Umask 0x0000FF00
#define Vmask 0x000000FF
#define trY   0x00300000
#define trU   0x00000700
#define trV   0x00000006

static inline int yuv_diff(uint32_t yuv1, uint32_t yuv2)
{
#if defined(ARMV6)
    register uint32_t tmp1, tmp2;

    asm(
        "usub8 %[tmp1], %[value1], %[value2]   \n\t" // tmp1 = value1 - value2             // tmp1 = yuv1 - yuv2
        "usub8 %[value2], %[value2], %[value1] \n\t" // value2 = value2 - value1           // value2 = yuv2 - yuv1
        "mov %[tmp2], #0                       \n\t" // tmp2 = 0
        "movw %[value1], #0x0706               \n\t"
        "movt %[value1], #0x30                 \n\t" // value1 = 0x300706
        "sel %[tmp1], %[value2], %[tmp1]       \n\t" // tmp1 = (value2 >= 0)?value2:tmp1   // tmp1 = abs(yuv1 - yuv2)
        "usub8 %[value2], %[value1], %[tmp1]   \n\t" // value2 = value1 - tmp1             // value2 = 0x300706 - abs(yuv1 - yuv2)
        "sel %[value1], %[tmp2], %[value1]         " // value1 = (value2 >= 0)?tmp2:value1 // value1 = (0x300706 >= abs(yuv1 - yuv2))?0:0x300706

        : [value1] "+r" (yuv1), [value2] "+r" (yuv2), [tmp1] "=r" (tmp1), [tmp2] "=r" (tmp2)
        :
        : "cc"
    );

    return yuv1;
#else
    return ( ( abs((yuv1 & Ymask) - (yuv2 & Ymask)) > trY ) ||
            ( abs((yuv1 & Umask) - (yuv2 & Umask)) > trU ) ||
            ( abs((yuv1 & Vmask) - (yuv2 & Vmask)) > trV ) );
#endif
}


#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_BGR565))
    #define G_SZ (6)
    #define G_MASK (63 << 2)
#elif ((HQX_FORMAT_16 == HQX_FORMAT_RGB555) || (HQX_FORMAT_16 == HQX_FORMAT_BGR555))
    #define G_SZ (5)
    #define G_MASK (31 << 3)
#else
    #error Bad format selected - HQX_FORMAT_16
#endif

#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_RGB555))
    #define RB_0 r
    #define RB_2 b
#else
    #define RB_0 b
    #define RB_2 r
#endif

#define G_SH (G_SZ - 3)
#define RB_SH_2 (G_SZ + 2)

#if !defined(ARMV7)
#ifdef __cplusplus
extern "C"
#endif
void convert_yuv_16(const uint16_t *src, uint32_t *dst, unsigned int width)
{
    unsigned int index;

    for (index = 0; index < width; index++)
    {
        unsigned int rgbvalue;
        int r, g, b;
        unsigned int Y, u, v;

        rgbvalue = src[index];

        RB_0 = (rgbvalue << 3      ) & (31 << 3);
        g    = (rgbvalue >> G_SH   ) & G_MASK;
        RB_2 = (rgbvalue >> RB_SH_2) & (31 << 3);

        RB_0 |= RB_0 >> 5;
        g    |= g    >> G_SZ;
        RB_2 |= RB_2 >> 5;

        Y = (r + g + b) >> 2;
        u = 128 + ((r - b) >> 2);
        v = 128 + ((-r + 2*g -b)>>3);

        dst[index] = (Y << 16) + (u << 8) + v;
    }

    dst[-1] = dst[0];
    dst[width] = dst[width - 1];
}

#ifdef __cplusplus
extern "C"
#endif
void convert_yuv_8_rgb_16(const uint8_t *src, uint32_t *yuvdst, uint16_t *rgbdst, const uint32_t *palette, unsigned int width)
{
    unsigned int index;

    for (index = 0; index < width; index++)
    {
        unsigned int rgbvalue;
        int r, g, b;
        unsigned int Y, u, v;

        rgbvalue = palette[src[index]];
        rgbdst[index] = rgbvalue;

        RB_0 = (rgbvalue << 3      ) & (31 << 3);
        g    = (rgbvalue >> G_SH   ) & G_MASK;
        RB_2 = (rgbvalue >> RB_SH_2) & (31 << 3);

        RB_0 |= RB_0 >> 5;
        g    |= g    >> G_SZ;
        RB_2 |= RB_2 >> 5;

        Y = (r + g + b) >> 2;
        u = 128 + ((r - b) >> 2);
        v = 128 + ((-r + 2*g -b)>>3);

        yuvdst[index] = (Y << 16) + (u << 8) + v;
    }

    yuvdst[-1] = yuvdst[0];
    yuvdst[width] = yuvdst[width - 1];
}
#endif

#ifdef __cplusplus
extern "C"
#endif
void calculate_pattern(const uint32_t *yuvsrc1, const uint32_t *yuvsrc2, const uint32_t *yuvsrc3, uint8_t *dst, unsigned int width)
{
    uint32_t e, newb, newe, newh;
    uint8_t pattern;

    // before first
    {
        newb = yuvsrc1[-1];
        newe = yuvsrc2[-1];
        newh = yuvsrc3[-1];
        pattern = 0;
        if (yuv_diff(newe, newb    )) pattern |= (1 << 0) | (1 << 1) | (1 << 2);
        if (yuv_diff(newe, newh    )) pattern |= (1 << 5) | (1 << 6) | (1 << 7);

        dst[-1] = pattern;
    }

    for (; width != 0; width--)
    {
        e = newe;
        pattern = 0;
        if (yuv_diff(e, yuvsrc1[-1])) pattern |= 1 << 0;
        if (yuv_diff(e, newb       )) pattern |= 1 << 1;
        newb = yuvsrc1[ 1];
        if (yuv_diff(e, newb       )) pattern |= 1 << 2;
        if (yuv_diff(e, yuvsrc2[-1])) pattern |= 1 << 3;
        newe = yuvsrc2[ 1];
        if (yuv_diff(e, newe       )) pattern |= 1 << 4;
        if (yuv_diff(e, yuvsrc3[-1])) pattern |= 1 << 5;
        if (yuv_diff(e, newh       )) pattern |= 1 << 6;
        newh = yuvsrc3[ 1];
        if (yuv_diff(e, newh       )) pattern |= 1 << 7;

        *dst = pattern;

        yuvsrc1++;
        yuvsrc2++;
        yuvsrc3++;
        dst++;
    }

    // after last
    {
        pattern = 0;
        if (yuv_diff(newe, newb    )) pattern |= (1 << 0) | (1 << 1) | (1 << 2);
        if (yuv_diff(newe, newh    )) pattern |= (1 << 5) | (1 << 6) | (1 << 7);

        *dst = pattern;
    }
}

